// Yet Another Units Library (YANL) - A C++ library for zero-overhead
// dimensional analysis and unit manipulation/conversion.
// 
// Copyright (C) 2003 Matthias Christian Schabel
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#ifndef MCS_CGS_UNITS_HPP
#define MCS_CGS_UNITS_HPP

#include <string>

#include <mcs/units/dimensioned_quantity.hpp>
#include <mcs/units/physical_units.hpp>

namespace mcs {

namespace units { 

class SIModel;

// CGS units
struct CGSModel :
    public PhysicalUnitModel<CGSModel>
{ };

template<> 
struct UnitInfo<CGSModel,length_tag> 
{ 
    typedef SIModel  base_type;

	static std::string symbol()				{ return "centimeter"; }
	static std::string abbreviation()		{ return "cm"; }
	
	template<typename Y> static Y toBase()	{ return Y(0.01); }
};
	
template<> 
struct UnitInfo<CGSModel,mass_tag> 
{ 
    typedef SIModel  base_type;

	static std::string symbol()				{ return "gram"; }
	static std::string abbreviation()		{ return "g"; }
	
	template<typename Y> static Y toBase()	{ return Y(0.001); }
};
	
template<> 
struct UnitInfo<CGSModel,time_tag> 
{ 
    typedef SIModel  base_type;

	static std::string symbol()				{ return "second"; }
	static std::string abbreviation()		{ return "s"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<CGSModel,current_tag> 
{ 
    typedef SIModel  base_type;

	static std::string symbol()				{ return "biot"; }
	static std::string abbreviation()		{ return "Bi"; }
	
	template<typename Y> static Y toBase()	{ return Y(10); }
};
	
template<> 
struct UnitInfo<CGSModel,temperature_tag> 
{ 
    typedef SIModel  base_type;

	static std::string symbol()				{ return "kelvin"; }
	static std::string abbreviation()		{ return "K"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<CGSModel,quantity_tag> 
{ 
    typedef SIModel  base_type;

	static std::string symbol()				{ return "mole"; }
	static std::string abbreviation()		{ return "mol"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<CGSModel,intensity_tag> 
{ 
    typedef SIModel  base_type;

	static std::string symbol()				{ return "candela"; }
	static std::string abbreviation()		{ return "cd"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};

template<> 
struct UnitInfo<CGSModel,angle_tag> 
{ 
    typedef SIModel  base_type;

	static std::string symbol()				{ return "radian"; }
	static std::string abbreviation()		{ return "rd"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<CGSModel,solid_angle_tag> 
{ 
    typedef SIModel  base_type;

	static std::string symbol()				{ return "steradian"; }
	static std::string abbreviation()		{ return "sr"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
namespace cgs {

static const CGSModel::length_unit					_centimeter,
													_centimeters;
static const CGSModel::mass_unit					_gram,
													_grams;
static const CGSModel::time_unit					_second,
													_seconds;
static const CGSModel::current_unit					_biot,
													_biots;
static const CGSModel::temperature_unit				_kelvin,
													_kelvins;
static const CGSModel::quantity_unit				_mole,
													_moles;
static const CGSModel::intensity_unit				_candela,
													_candelas;
static const CGSModel::angle_unit					_radian,
													_radians;
static const CGSModel::solid_angle_unit				_steradian,
													_steradians;													
												
static const CGSModel::acceleration_unit			_gal,
													_gals;
static const CGSModel::area_unit					_square_centimeter,
													_square_centimeters;
static const CGSModel::energy_unit					_erg,
													_ergs;
static const CGSModel::force_unit					_dyne,
													_dynes;
static const CGSModel::frequency_unit				_hertz;
static const CGSModel::pressure_unit				_barye,
													_baryes;
static const CGSModel::velocity_unit				_centimeter_per_second,
													_centimeters_per_second;
static const CGSModel::volume_unit					_cubic_centimeter,
													_cubic_centimeters;
static const CGSModel::wavenumber_unit				_kayser,
													_kaysers;
														
static const CGSModel::capacitance_unit				_abfarad,
													_abfarads;
static const CGSModel::conductance_unit				_abmho,
													_abmhos;
static const CGSModel::electric_charge_unit			_abcoulomb,
													_abcoulombs;
static const CGSModel::electric_potential_unit		_abvolt,
													_abvolts;
static const CGSModel::magnetic_flux_unit			_maxwell,
													_maxwells;
static const CGSModel::magnetic_flux_density_unit	_gauss;
static const CGSModel::inductance_unit				_abhenry,
													_abhenrys;
static const CGSModel::resistance_unit				_abohm,
													_abohms;

} // namespace cgs

template<class Y>
struct CGS : 
	public PhysicalQuantityModel<Y,CGSModel>
{ };

} // namespace units

} // namespace mcs

#endif // MCS_CGS_UNITS_HPP
