// Yet Another Units Library (YANL) - A C++ library for zero-overhead
// dimensional analysis and unit manipulation/conversion.
// 
// Copyright (C) 2003 Matthias Christian Schabel
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#ifndef MCS_DIMENSIONED_IO_HPP
#define MCS_DIMENSIONED_IO_HPP

#include <iostream>
#include <string>
#include <utility>
#include <vector>

#include <boost/mpl/for_each.hpp>

#include <mcs/units/dimensioned_quantity.hpp>

namespace mcs {

namespace units { 

// output formatting utility class
class rational : 
    public std::pair<long,long>
{
    public:
        template<long N,long D>
        rational(Rational<N,D> r) : 
            std::pair<long,long>(r.numerator(),r.denominator()) 
        { }
    
        long numerator() const      { return this->first; }
        long denominator() const    { return this->second; }
        
    private:
        rational();
};

std::ostream& operator<<(std::ostream& os,const rational& val)
{
    if (val.denominator() == 1)
    {
        if (val.numerator() != 1) 
            os << val.numerator();
    }
    else if (val.numerator() == val.denominator())
    {
        os << "1";
    }
    else
    {
        os << "(" << val.numerator() << "/" << val.denominator() << ")";
    }
        
    return os;
}

namespace {

template<class D>
inline void 
abbreviationFunc(std::vector<std::string>& avec,const std::string& str)
{
    avec.push_back(str);
}

template<class D>
inline void 
exponentFunc(std::vector<rational>& evec,const rational& val)
{
    evec.push_back(val);
}

template<class Model>
class DimensionedUnitOutputHelper
{
    public:
        DimensionedUnitOutputHelper(std::vector<std::string>& avec,
                                    std::vector<rational>& rvec) : 
            avec_(avec),
            rvec_(rvec)
        { }
    
        template<class D>
        void operator()(D) const
        {
            typedef typename get_dim<D>::type    dim_type;
            typedef typename get_value<D>::type    value_type;
            
            typedef UnitInfo<Model,dim_type>    m_type;
    
            abbreviationFunc<D>(avec_,m_type::abbreviation());
            exponentFunc<D>(rvec_,rational(value_type()));
        }
    
    private:
        std::vector<std::string>&    avec_;
        std::vector<rational>&       rvec_;
};

} // namespace

template<class Model,class Dim>
std::ostream& operator<<(std::ostream& os,const DimensionedUnit<Model,Dim>&)
{
    std::vector<std::string>    avec;
    std::vector<rational>       rvec;
    
    DimensionedUnitOutputHelper<Model>    helper(avec,rvec);
    
    if (boost::is_same<Dim,dimensionless_type>::value)
    {
        os << "dimensionless ";
    }
    else
    {
        boost::mpl::for_each<Dim>(helper);

        for (std::size_t i=0;i<avec.size();++i)
        {
            os << avec[i];
            
            if (rvec[i].numerator() != rvec[i].denominator())
                os << "^";
                
            os << rvec[i];
            
            if (i != avec.size()-1)
                os << " ";
        }
    }
    
    return os;
}

template<class Y,class Model,class Dim>
std::ostream& operator<<
(std::ostream& os,
 const DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >& val)
{
    const DimensionedUnit<Model,Dim>    unit;
    
    os << val.value() << " " << unit;

    return os;
}

} // namespace units

} // namespace mcs

#endif // MCS_DIMENSIONED_IO_HPP
