// Yet Another Units Library (YANL) - A C++ library for zero-overhead
// dimensional analysis and unit manipulation/conversion.
// 
// Copyright (C) 2003 Matthias Christian Schabel
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#ifndef MCS_DIMENSIONED_QUANTITY_HPP
#define MCS_DIMENSIONED_QUANTITY_HPP

#include <cmath>
#include <vector>

#include <boost/static_assert.hpp>
#include <boost/type_traits.hpp>

#include <boost/mpl/at.hpp>
#include <boost/mpl/for_each.hpp>

#include <mcs/units/dimensioned_unit.hpp>

namespace mcs {

namespace units {

template<typename Model,typename Tag> struct UnitInfo;

template<class Y,class T> class DimensionedQuantity;

template<class Y,class Model,class Dim>
class DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
{
    public:
        typedef DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >  this_type;
        
        typedef Y                                             value_type;
        typedef Model                                         model_type;
        typedef Dim                                           dimension_type;
        typedef DimensionedUnit<model_type,dimension_type>    unit_type;
         
        DimensionedQuantity() : val_() { }
        explicit DimensionedQuantity(value_type val) : val_(val) { }
        DimensionedQuantity(const this_type& source) : val_(source.val_) { }

        // explicit conversions
        template<class YY>
        explicit 
        DimensionedQuantity
        (const DimensionedQuantity<YY,DimensionedUnit<model_type,dimension_type> >& source) : 
            val_(static_cast<value_type>(source.value())) 
        { }
        
        // implicit conversion between commensurate dimension lists is allowed
        template<class Dim2> 
        DimensionedQuantity
        (const DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> >& source)
        {
            // ensure dimension lists are commensurate
            BOOST_STATIC_ASSERT((is_commensurate<Dim,Dim2>::value == true));
            
            val_ = source.value();
        }

        // conversion between different models must be explicit
        template<class Model2,class Dim2> 
        explicit DimensionedQuantity
        (const DimensionedQuantity<Y,DimensionedUnit<Model2,Dim2> >&);

        ~DimensionedQuantity() { }
        
        this_type& operator=(const this_type& source) 
        { 
             if (this != &source) 
             { 
                 val_ = source.val_; 
             } 
             
             return *this; 
        }
        
        // can multiply or divide by scalar
        this_type& operator*=(value_type val)            
        { val_ *= val; return *this; }
        
        this_type& operator/=(value_type val)            
        { val_ /= val; return *this; }
        
        // can add or subtract same unit type
        this_type& operator+=(this_type source)            
        { val_ += source.val_; return *this; }
        
        this_type& operator-=(this_type source)            
        { val_ -= source.val_; return *this; }

        value_type& value()                             
        { return val_; }
        
        value_type value() const                        
        { return val_; }
        
    private:
        value_type    val_;
};

namespace {

template<class Y,class D>
inline void 
conversionFunc(std::vector<Y>& fac,Y val)
{
    fac.push_back(val);
}

template<class Y,
         class Model1,
         class Model2>
class DimensionedQuantityConversionHelper
{
    public:
        DimensionedQuantityConversionHelper(std::vector<Y>& fac) : 
            fac_(fac) 
        { }
    
        template<class D>
        void operator()(D) const
        {
            typedef typename get_dim<D>::type      dim_type;
            typedef typename get_value<D>::type    value_type;
            
            typedef UnitInfo<Model1,dim_type>    m1_type;
            typedef UnitInfo<Model2,dim_type>    m2_type;
            
            // make sure Models are convertible
            BOOST_STATIC_ASSERT((boost::is_same<
                                                typename m1_type::base_type,
                                                typename m2_type::base_type
                                               >::value == true));

            const Y ex = value<Y>(value_type()),
                    cv1 = m1_type::template toBase<Y>(),
                    cv2 = m2_type::template toBase<Y>(),
                    val = std::pow(cv2/cv1,ex);
    
            conversionFunc<Y,D>(fac_,val);
        }
    
    private:
        std::vector<Y>&    fac_;
};

} // namespace

template<class Y,
         class Model,
         class Dim>
template<class Model2,class Dim2>
DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >::DimensionedQuantity
(const DimensionedQuantity<Y,DimensionedUnit<Model2,Dim2> >& source)
{
    // ensure dimension lists are commensurate
    BOOST_STATIC_ASSERT((is_commensurate<Dim,Dim2>::value == true));

    val_ = source.value();
    
    static bool          initialized = false;
    static const long    sz = boost::mpl::size<Dim>::value;
    
    if (sz > 0)
    {
        static std::vector<Y>    fac;
        
        if (!initialized)
        {
            fac.reserve(sz);
            
            DimensionedQuantityConversionHelper<Y,Model,Model2>  helper(fac);
            
            boost::mpl::for_each<Dim>(helper);
            
            initialized = true;
        }
            
        for (std::size_t i=0;i<fac.size();++i)
        { 
            val_ *= fac[i];
        }
    }
}

// specialization for dimensionless quantities -> implicit convertions to Y
template<class Y,
         class Model>
class DimensionedQuantity<Y,DimensionedUnit<Model,dimensionless_type> >
{
    public:
        typedef DimensionedQuantity<
                                    Y,
                                    DimensionedUnit<Model,dimensionless_type> 
                                   >    this_type;
                                   
        typedef Y                                             value_type;
        typedef Model                                         model_type;
        typedef dimensionless_type                            dimension_type;
        typedef DimensionedUnit<model_type,dimension_type>    unit_type;
         
        DimensionedQuantity() : val_() { }
        DimensionedQuantity(value_type val) : val_(val) { }
        DimensionedQuantity(const this_type& source) : val_(source.val_) { }
        ~DimensionedQuantity() { }
        
        template<class YY>
        explicit 
        DimensionedQuantity
        (const DimensionedQuantity<
                                   YY,
                                   DimensionedUnit<Model,dimension_type> 
                                  >& source) : 
            val_(static_cast<value_type>(source.value())) 
        { }
        
        // unit conversion must be explicit
        template<class Model2> 
        explicit 
        DimensionedQuantity
        (const DimensionedQuantity<
                                   Y,
                                   DimensionedUnit<Model2,dimensionless_type> 
                                  >& source) :
            val_(source.value()) 
        { }

        this_type& operator=(const this_type& source) 
        { 
            if (this != &source) 
                val_ = source.val_; 
                
            return *this; 
        }
        
        operator value_type() const    { return val_; }
        
        value_type& value()            { return val_; }
        value_type value() const       { return val_; }
        
    private:
        value_type    val_;
};

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> >
operator+(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > lhs,
          DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > rhs)
{
    // ensure dimension lists are commensurate
    BOOST_STATIC_ASSERT((is_commensurate<Dim1,Dim2>::value == true));

    return 
    DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> >
    (lhs.value()+rhs.value());
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> >
operator-(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > lhs,
          DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > rhs)
{
    // ensure dimension lists are commensurate
    BOOST_STATIC_ASSERT((is_commensurate<Dim1,Dim2>::value == true));

    return 
    DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> >
    (lhs.value()-rhs.value());
}

template<class Y,
         class Model,
         class Dim>
DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
operator*(DimensionedUnit<Model,Dim>,Y rhs)
{
    return 
    DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
    (rhs);
}

template<class Y,
         class Model,
         class Dim>
DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
operator/(DimensionedUnit<Model,Dim>,Y rhs)
{
    return 
    DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
    (static_cast<Y>(1)/rhs);
}

template<class Y,
         class Model,
         class Dim>
DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
operator*(Y lhs,DimensionedUnit<Model,Dim>)
{
    return 
    DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
    (lhs);
}

template<class Y,
         class Model,
         class Dim>
DimensionedQuantity<Y,DimensionedUnit<Model,typename Negate<Dim>::type> >
operator/(Y lhs,DimensionedUnit<Model,Dim>)
{
    return 
    DimensionedQuantity<Y,DimensionedUnit<Model,typename Negate<Dim>::type> >
    (lhs);
}

template<class Y,
         class Model,
         class Dim>
DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
operator*(DimensionedQuantity<Y,DimensionedUnit<Model,Dim> > lhs,Y rhs)
{
    return 
    DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
    (lhs.value()*rhs);
}

template<class Y,
         class Model,
         class Dim>
DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
operator/(DimensionedQuantity<Y,DimensionedUnit<Model,Dim> > lhs,Y rhs)
{
    return 
    DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
    (lhs.value()/rhs);
}

template<class Y,
         class Model,
         class Dim>
DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
operator*(Y lhs,DimensionedQuantity<Y,DimensionedUnit<Model,Dim> > rhs)
{
    return 
    DimensionedQuantity<Y,DimensionedUnit<Model,Dim> >
    (lhs*rhs.value());
}

template<class Y,
         class Model,
         class Dim>
DimensionedQuantity<Y,DimensionedUnit<Model,typename Negate<Dim>::type> >
operator/(Y lhs,DimensionedQuantity<Y,DimensionedUnit<Model,Dim> > rhs)
{
    return 
    DimensionedQuantity<Y,DimensionedUnit<Model,typename Negate<Dim>::type> >
    (lhs/rhs.value());
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
DimensionedQuantity<
                    Y,
                    DimensionedUnit<Model,typename Multiply<Dim1,Dim2>::type> 
                   >
operator*(DimensionedUnit<Model,Dim1>,
          DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > rhs)
{
    return 
    DimensionedQuantity<
                        Y,
                        DimensionedUnit<
                                        Model,
                                        typename Multiply<Dim1,Dim2>::type
                                       > 
                       >
    (rhs.value());
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
DimensionedQuantity<
                    Y,
                    DimensionedUnit<Model,typename Divide<Dim1,Dim2>::type> 
                   >
operator/(DimensionedUnit<Model,Dim1>,
          DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > rhs)
{
    return 
    DimensionedQuantity<
                        Y,
                        DimensionedUnit<
                                        Model,
                                        typename Divide<Dim1,Dim2>::type
                                       > 
                       >
    (Y(1)/rhs.value());
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
DimensionedQuantity<
                    Y,
                    DimensionedUnit<Model,typename Multiply<Dim1,Dim2>::type> 
                   >
operator*(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > lhs,
          DimensionedUnit<Model,Dim2>)
{
    return 
    DimensionedQuantity<
                        Y,
                        DimensionedUnit<
                                        Model,
                                        typename Multiply<Dim1,Dim2>::type
                                       > 
                       >
    (lhs.value());
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
DimensionedQuantity<
                    Y,
                    DimensionedUnit<Model,typename Divide<Dim1,Dim2>::type> 
                   >
operator/(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > lhs,
          DimensionedUnit<Model,Dim2>)
{
    return 
    DimensionedQuantity<
                        Y,
                        DimensionedUnit<
                                        Model,
                                        typename Divide<Dim1,Dim2>::type
                                       > 
                       >
    (lhs.value());
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
DimensionedQuantity<
                    Y,
                    DimensionedUnit<Model,typename Multiply<Dim1,Dim2>::type> 
                   >
operator*(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > lhs,
          DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > rhs)
{
    return 
    DimensionedQuantity<
                        Y,
                        DimensionedUnit<
                                        Model,
                                        typename Multiply<Dim1,Dim2>::type
                                       > 
                       >
    (lhs.value()*rhs.value());
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
DimensionedQuantity<
                    Y,
                    DimensionedUnit<Model,typename Divide<Dim1,Dim2>::type> 
                   >
operator/(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > lhs,
          DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > rhs)
{
    return 
    DimensionedQuantity<
                        Y,
                        DimensionedUnit<
                                        Model,
                                        typename Divide<Dim1,Dim2>::type
                                       > 
                       >
    (lhs.value()/rhs.value());
}

// rational powers and roots
template<class Rat,class Y,class Model,class Dim>
DimensionedQuantity<Y,DimensionedUnit<Model,typename Power<Dim,Rat>::type> >
power(DimensionedQuantity<Y,DimensionedUnit<Model,Dim> > val)
{ 
    return 
    DimensionedQuantity<
                        Y,
                        DimensionedUnit<Model,typename Power<Dim,Rat>::type> 
                       >
    (std::pow(val.value(),value<Y>(Rat())));
}

template<long N,class Y,class Model,class Dim>
DimensionedQuantity<
                    Y,
                    DimensionedUnit<
                                    Model,
                                    typename Power<Dim,Rational<N> >::type
                                   > 
                   >
power(DimensionedQuantity<Y,DimensionedUnit<Model,Dim> > val)
{ 
    return 
    DimensionedQuantity<
                        Y,
                        DimensionedUnit<
                                        Model,
                                        typename Power<Dim,Rational<N> >::type
                                       > 
                       >(std::pow(val.value(),value<Y>(Rational<N>())));
}

template<class Rat,class Y,class Model,class Dim>
DimensionedQuantity<
                    Y,
                    DimensionedUnit<Model,typename Root<Dim,Rat>::type> 
                   >
root(DimensionedQuantity<Y,DimensionedUnit<Model,Dim> > val)
{ 
    return 
    DimensionedQuantity<
                        Y,
                        DimensionedUnit<Model,typename Root<Dim,Rat>::type> 
                       >
    (std::pow(val.value(),Y(1)/value<Y>(Rat())));
}

template<long N,class Y,class Model,class Dim>
DimensionedQuantity<
                    Y,
                    DimensionedUnit<
                                    Model,
                                    typename Root<Dim,Rational<N> >::type
                                   > 
                   >
root(DimensionedQuantity<Y,DimensionedUnit<Model,Dim> > val)
{ 
    return 
    DimensionedQuantity<
                        Y,
                        DimensionedUnit<
                                        Model,
                                        typename Root<Dim,Rational<N> >::type
                                       > 
                       >
    (std::pow(val.value(),Y(1)/value<Y>(Rational<N>())));
}

// comparison operators between same model
template<class Y,
         class Model,
         class Dim1,
         class Dim2>
bool 
operator==(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > val1,
           DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > val2)
{
    // ensure dimension lists are commensurate
    BOOST_STATIC_ASSERT((is_commensurate<Dim1,Dim2>::value == true));

    return val1.value() == val2.value();
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
bool 
operator!=(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > val1,
           DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > val2)
{
    // ensure dimension lists are commensurate
    BOOST_STATIC_ASSERT((is_commensurate<Dim1,Dim2>::value == true));

return val1.value() != val2.value();
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
bool 
operator<(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > val1,
          DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > val2)
{
    // ensure dimension lists are commensurate
    BOOST_STATIC_ASSERT((is_commensurate<Dim1,Dim2>::value == true));

    return val1.value() < val2.value();
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
bool 
operator<=(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > val1,
           DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > val2)
{
    // ensure dimension lists are commensurate
    BOOST_STATIC_ASSERT((is_commensurate<Dim1,Dim2>::value == true));

    return val1.value() <= val2.value();
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
bool 
operator>(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > val1,
          DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > val2)
{
    // ensure dimension lists are commensurate
    BOOST_STATIC_ASSERT((is_commensurate<Dim1,Dim2>::value == true));

    return val1.value() > val2.value();
}

template<class Y,
         class Model,
         class Dim1,
         class Dim2>
bool 
operator>=(DimensionedQuantity<Y,DimensionedUnit<Model,Dim1> > val1,
           DimensionedQuantity<Y,DimensionedUnit<Model,Dim2> > val2)
{
    // ensure dimension lists are commensurate
    BOOST_STATIC_ASSERT((is_commensurate<Dim1,Dim2>::value == true));

    return val1.value() >= val2.value();
}

} // namespace units

} // namespace mcs

#endif // MCS_DIMENSIONED_QUANTITY_HPP
