// Yet Another Units Library (YANL) - A C++ library for zero-overhead
// dimensional analysis and unit manipulation/conversion.
// 
// Copyright (C) 2003 Matthias Christian Schabel
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#ifndef MCS_DIMENSIONED_UNIT_HPP
#define MCS_DIMENSIONED_UNIT_HPP

#include <mcs/units/dimension.hpp>

namespace mcs {

namespace units { 

// class representing a model-dependent unit type (e.g. meters, Kelvin, etc...)
//     with no associated value.
template<class Model,
         class Dim>
class DimensionedUnit
{
    public:
        typedef DimensionedUnit<Model,Dim>   this_type;
        typedef Model                        model_type;
        typedef Dim                          dimension_type; 
        
        DimensionedUnit() { }
        DimensionedUnit(const this_type&) { }
        ~DimensionedUnit() { }
        
        this_type& operator=(const this_type&) { }
};

// multiply two unit types
template<class Model,
         class Dim1,
         class Dim2>
struct Multiply< DimensionedUnit<Model,Dim1>,DimensionedUnit<Model,Dim2> >
{
    typedef DimensionedUnit<Model,typename Multiply<Dim1,Dim2>::type>    type;
};

// divide two unit types 
template<class Model,
         class Dim1,
         class Dim2>
struct Divide< DimensionedUnit<Model,Dim1>,DimensionedUnit<Model,Dim2> >
{
    typedef DimensionedUnit<Model,typename Divide<Dim1,Dim2>::type>    type;
};

// runtime multiply two unit type variables
template<class Model,
         class Dim1,
         class Dim2>
typename Multiply< DimensionedUnit<Model,Dim1>,
                   DimensionedUnit<Model,Dim2> >::type
operator*(DimensionedUnit<Model,Dim1>,DimensionedUnit<Model,Dim2>)
{
    typedef typename Multiply< DimensionedUnit<Model,Dim1>,
                               DimensionedUnit<Model,Dim2> >::type    type;
    
    return type();
}

// runtime divide two unit type variables
template<class Model,
         class Dim1,
         class Dim2>
typename Divide< DimensionedUnit<Model,Dim1>,
                 DimensionedUnit<Model,Dim2> >::type
operator/(DimensionedUnit<Model,Dim1>,DimensionedUnit<Model,Dim2>)
{
    typedef typename Divide< DimensionedUnit<Model,Dim1>,
                             DimensionedUnit<Model,Dim2> >::type    type;
    
    return type();
}

// rational powers and roots of unit type variables
//
//    e.g. power<Rational<4,3> >(_meter) -> _meter^(4/3)
//
template<class Rat,class Model,class Dim>
DimensionedUnit<Model,typename Power<Dim,Rat>::type>
power(DimensionedUnit<Model,Dim>)
{ 
    return DimensionedUnit<Model,typename Power<Dim,Rat>::type>();
}

template<long N,class Model,class Dim>
DimensionedUnit<Model,typename Power<Dim,Rational<N> >::type>
power(DimensionedUnit<Model,Dim>)
{ 
    return DimensionedUnit<Model,typename Power<Dim,Rational<N> >::type>();
}

template<class Rat,class Model,class Dim>
DimensionedUnit<Model,typename Root<Dim,Rat>::type>
root(DimensionedUnit<Model,Dim>)
{ 
    return DimensionedUnit<Model,typename Root<Dim,Rat>::type>();
}

template<long N,class Model,class Dim>
DimensionedUnit<Model,typename Root<Dim,Rational<N> >::type>
root(DimensionedUnit<Model,Dim>)
{ 
    return DimensionedUnit<Model,typename Root<Dim,Rational<N> >::type>();
}

} // namespace units

} // namespace mcs

#endif // MCS_DIMENSIONED_UNIT_HPP
