// Yet Another Units Library (YANL) - A C++ library for zero-overhead
// dimensional analysis and unit manipulation/conversion.
// 
// Copyright (C) 2003 Matthias Christian Schabel
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#ifndef MCS_MEASUREMENT_HPP
#define MCS_MEASUREMENT_HPP

#include <cmath>
#include <iostream>

namespace mcs {

namespace {

template<class Y>
Y sqr(Y val)
{ return val*val; }

} // namespace

template<class Y>
class Measurement
{    
    public:
        typedef Measurement<Y>    this_type;
        typedef Y                value_type;
        
        Measurement(value_type val,value_type err = Y()) : 
            value_(val),
            error_(std::abs(err)) 
        { }
        
        Measurement(const this_type& source) : 
            value_(source.value_),
            error_(source.error_) 
        { }
        
        value_type value() const                        { return value_; }
        value_type error() const                        { return error_; }
        
        this_type& operator+=(value_type val)            
        { 
            value_ += val; 
            return *this; 
        }
        
        this_type& operator-=(value_type val)            
        { 
            value_ -= val; 
            return *this; 
        }
        
        this_type& operator*=(value_type val)            
        { 
            value_ *= val; 
            error_ *= val; 
            return *this; 
        }
        
        this_type& operator/=(value_type val)            
        { 
            value_ /= val; 
            error_ /= val; 
            return *this; 
        }
        
        this_type& operator+=(const this_type& source);
        this_type& operator-=(const this_type& source);        
        this_type& operator*=(const this_type& source);        
        this_type& operator/=(const this_type& source);

    private:
        value_type  value_,
                    error_;
};

template<class Y>
Measurement<Y>&
Measurement<Y>::operator+=(const this_type& source)
{
    error_ = std::sqrt(sqr(error_)+sqr(source.error_));
    value_ += source.value_;
    
    return *this;
}

template<class Y>
Measurement<Y>&
Measurement<Y>::operator-=(const this_type& source)
{
    error_ = std::sqrt(sqr(error_)+sqr(source.error_));
    value_ -= source.value_;
    
    return *this;
}

template<class Y>
Measurement<Y>&
Measurement<Y>::operator*=(const this_type& source)
{
    error_ = (value_*source.value_)*
              std::sqrt(sqr(error_/value_)+
                        sqr(source.error_/source.value_));
    value_ *= source.value_;
    
    return *this;
}

template<class Y>
Measurement<Y>&
Measurement<Y>::operator/=(const this_type& source)
{
    error_ = (value_/source.value_)*
              std::sqrt(sqr(error_/value_)+
                        sqr(source.error_/source.value_));
    value_ /= source.value_;
    
    return *this;
}

// non-member operators
template<class Y>
std::ostream& operator<<(std::ostream& os,const Measurement<Y>& val)
{
    os << val.value() << "(+/-" << val.error() << ")";
    
    return os;
}

// value_type op Measurement
template<class Y>
Measurement<Y>
operator+(Y lhs,
          const Measurement<Y>& rhs)
{
    return (Measurement<Y>(lhs,Y(0))+=rhs);
}

template<class Y>
Measurement<Y>
operator-(Y lhs,
          const Measurement<Y>& rhs)
{
    return (Measurement<Y>(lhs,Y(0))-=rhs);
}

template<class Y>
Measurement<Y>
operator*(Y lhs,
          const Measurement<Y>& rhs)
{
    return (Measurement<Y>(lhs,Y(0))*=rhs);
}

template<class Y>
Measurement<Y>
operator/(Y lhs,
          const Measurement<Y>& rhs)
{
    return (Measurement<Y>(lhs,Y(0))/=rhs);
}

// Measurement op value_type
template<class Y>
Measurement<Y>
operator+(const Measurement<Y>& lhs,
          Y rhs)
{
    return (Measurement<Y>(lhs)+=Measurement<Y>(rhs,Y(0)));
}

template<class Y>
Measurement<Y>
operator-(const Measurement<Y>& lhs,
          Y rhs)
{
    return (Measurement<Y>(lhs)-=Measurement<Y>(rhs,Y(0)));
}

template<class Y>
Measurement<Y>
operator*(const Measurement<Y>& lhs,
          Y rhs)
{
    return (Measurement<Y>(lhs)*=Measurement<Y>(rhs,Y(0)));
}

template<class Y>
Measurement<Y>
operator/(const Measurement<Y>& lhs,
          Y rhs)
{
    return (Measurement<Y>(lhs)/=Measurement<Y>(rhs,Y(0)));
}

// Measurement op Measurement
template<class Y>
Measurement<Y>
operator+(const Measurement<Y>& lhs,
          const Measurement<Y>& rhs)
{
    return (Measurement<Y>(lhs)+=rhs);
}

template<class Y>
Measurement<Y>
operator-(const Measurement<Y>& lhs,
          const Measurement<Y>& rhs)
{
    return (Measurement<Y>(lhs)-=rhs);
}

template<class Y>
Measurement<Y>
operator*(const Measurement<Y>& lhs,
          const Measurement<Y>& rhs)
{
    return (Measurement<Y>(lhs)*=rhs);
}

template<class Y>
Measurement<Y>
operator/(const Measurement<Y>& lhs,
          const Measurement<Y>& rhs)
{
    return (Measurement<Y>(lhs)/=rhs);
}

} // namespace mcs

namespace std {

// pow, exp, and log functions
template<class Y>
mcs::Measurement<Y>
pow(const mcs::Measurement<Y>& val,Y m)
{
    const Y newval = std::pow(val.value(),m),
            err = newval*std::sqrt(sqr(m*val.error()/val.value()));
    
    return mcs::Measurement<Y>(newval,err);
}

template<class Y>
mcs::Measurement<Y>
pow(const mcs::Measurement<Y>& val,const mcs::Measurement<Y>& m)
{
    const Y newval = std::pow(val.value(),m.value()),
            // N.B. not correct for uncertainty in exponent
            err = newval*std::sqrt(sqr(m.value()*val.error()/val.value()));
    
    return mcs::Measurement<Y>(newval,err);
}
/*
// not yet implemented
template<class Y>
Measurement<Y>
exp(const Measurement<Y>& val)
{
    return (Measurement<Y>(lhs)/=rhs);
}

template<class Y>
Measurement<Y>
log(const Measurement<Y>& val)
{
    return (Measurement<Y>(lhs)/=rhs);
}
*/
} // namespace std

#endif // MCS_MEASUREMENT_HPP
