// Yet Another Units Library (YANL) - A C++ library for zero-overhead
// dimensional analysis and unit manipulation/conversion.
// 
// Copyright (C) 2003 Matthias Christian Schabel
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#ifndef MCS_PHYSICAL_UNITS_HPP
#define MCS_PHYSICAL_UNITS_HPP

#include <boost/mpl/list.hpp>

#include <mcs/units/dimension.hpp>

namespace mcs {

namespace units { 

// shorthand definitions of dimensions corresponding to various physical unit types 

struct length_tag { };
struct mass_tag { };
struct time_tag { };
struct current_tag { };
struct temperature_tag { };
struct quantity_tag { };
struct intensity_tag { };
struct angle_tag { };
struct solid_angle_tag { };

// basic units
typedef 
boost::mpl::list<dim<length_tag,Rational<1> > >       length_type;

typedef 
boost::mpl::list<dim<mass_tag,Rational<1> > >         mass_type;

typedef 
boost::mpl::list<dim<time_tag,Rational<1> > >         time_type;

typedef 
boost::mpl::list<dim<current_tag,Rational<1> > >      current_type;

typedef 
boost::mpl::list<dim<temperature_tag,Rational<1> > >  temperature_type;

typedef 
boost::mpl::list<dim<quantity_tag,Rational<1> > >     quantity_type;

typedef 
boost::mpl::list<dim<intensity_tag,Rational<1> > >    intensity_type;

typedef 
boost::mpl::list<dim<angle_tag,Rational<1> > >        angle_type;

typedef 
boost::mpl::list<dim<solid_angle_tag,Rational<1> > >  solid_angle_type;

// common derived units
// l t^-2
typedef 
boost::mpl::list<dim<length_tag,Rational<1> >,
                 dim<time_tag,Rational<-2> > >        acceleration_type;  

// l^2                               
typedef 
boost::mpl::list<dim<length_tag,Rational<2> > >       area_type; 

// l^2 m t^-2
typedef 
boost::mpl::list<dim<length_tag,Rational<2> >,
                 dim<mass_tag,Rational<1> >,
                 dim<time_tag,Rational<-2> > >        energy_type;                

// l m t^-2
typedef 
boost::mpl::list<dim<length_tag,Rational<1> >,
	             dim<mass_tag,Rational<1> >,
	             dim<time_tag,Rational<-2> > >        force_type;                    

// t^-1
typedef 
boost::mpl::list<dim<time_tag,Rational<-1> > >        frequency_type;                

// l^-3 m
typedef 
boost::mpl::list<dim<length_tag,Rational<-3> >,
                 dim<mass_tag,Rational<1> > >         mass_density_type;            

// l m t^-1
typedef 
boost::mpl::list<dim<length_tag,Rational<1> >,
                 dim<mass_tag,Rational<1> >,
                 dim<time_tag,Rational<-1> > >        momentum_type;                

// l^2 m t^-3
typedef 
boost::mpl::list<dim<length_tag,Rational<2> >,
                 dim<mass_tag,Rational<1> >,
                 dim<time_tag,Rational<-3> > >        power_type;                    

// l^-1 m t^-2
typedef 
boost::mpl::list<dim<length_tag,Rational<-1> >,
                 dim<mass_tag,Rational<1> >,
                 dim<time_tag,Rational<-2> > >        pressure_type;                
// l t^-1
typedef 
boost::mpl::list<dim<length_tag,Rational<1> >,
                 dim<time_tag,Rational<-1> > >        velocity_type;                
// l^3
typedef 
boost::mpl::list<dim<length_tag,Rational<3> > >       volume_type;                

// l^-1
typedef 
boost::mpl::list<dim<length_tag,Rational<-1> > >      wavenumber_type;            

// electromagnetic derived units
// l^-2 m^-1 t^4 c^2
typedef 
boost::mpl::list<dim<length_tag,Rational<-2> >,
                 dim<mass_tag,Rational<-1> >,
                 dim<time_tag,Rational<4> >,
                 dim<current_tag,Rational<2> > >      capacitance_type;            

// l^-2 m^-1 t^3 c^2
typedef 
boost::mpl::list<dim<length_tag,Rational<-2> >,
                 dim<mass_tag,Rational<-1> >,
                 dim<time_tag,Rational<3> >,
                 dim<current_tag,Rational<2> > >      conductance_type;                    

// t^1 c^1
typedef 
boost::mpl::list<dim<time_tag,Rational<1> >,
                 dim<current_tag,Rational<1> > >      electric_charge_type;        

// l^2 m t^-3 c^-1
typedef 
boost::mpl::list<dim<length_tag,Rational<2> >,
                 dim<mass_tag,Rational<1> >,
                 dim<time_tag,Rational<-3> >,
                 dim<current_tag,Rational<-1> > >     electric_potential_type;        

// l^2 m t^-2 c^-2
typedef 
boost::mpl::list<dim<length_tag,Rational<2> >,
                 dim<mass_tag,Rational<1> >,
                 dim<time_tag,Rational<-2> >,
                 dim<current_tag,Rational<-2> > >     inductance_type;                

// l^2 m t^-2 c^-1
typedef 
boost::mpl::list<dim<length_tag,Rational<2> >,
                 dim<mass_tag,Rational<1> >,
                 dim<time_tag,Rational<-2> >,
                 dim<current_tag,Rational<-1> > >     magnetic_flux_type;            

// m t^-2 c^-1
typedef 
boost::mpl::list<dim<mass_tag,Rational<1> >,
                 dim<time_tag,Rational<-2> >,
                 dim<current_tag,Rational<-1> > >     magnetic_flux_density_type;        

// l^-3 m^-1 t^4 c^2
typedef 
boost::mpl::list<dim<length_tag,Rational<-3> >,
                 dim<mass_tag,Rational<-1> >,
                 dim<time_tag,Rational<4> >,
                 dim<current_tag,Rational<2> > >      permittivity_type;                

// l^2 m t^-3 c^-2  
typedef 
boost::mpl::list<dim<length_tag,Rational<2> >,
                 dim<mass_tag,Rational<1> >,
                 dim<time_tag,Rational<-3> >,
                 dim<current_tag,Rational<-2> > >     resistance_type;              

// miscellaneous
// t^-1 q^1
typedef 
boost::mpl::list<dim<time_tag,Rational<-1> >,
                 dim<quantity_tag,Rational<1> > >     catalytic_activity_type;    

// base class defining common units
template<class Model>
struct PhysicalUnitModel
{
	typedef DimensionedUnit<Model,length_type>					length_unit;
	typedef DimensionedUnit<Model,mass_type>					mass_unit;
	typedef DimensionedUnit<Model,time_type>					time_unit;
	typedef DimensionedUnit<Model,current_type>					current_unit;
	typedef DimensionedUnit<Model,temperature_type>				temperature_unit;
	typedef DimensionedUnit<Model,quantity_type>				quantity_unit;
	typedef DimensionedUnit<Model,intensity_type>				intensity_unit;
	typedef DimensionedUnit<Model,angle_type>					angle_unit;
	typedef DimensionedUnit<Model,solid_angle_type>				solid_angle_unit;

	typedef DimensionedUnit<Model,acceleration_type>			acceleration_unit;
	typedef DimensionedUnit<Model,area_type>					area_unit;
	typedef DimensionedUnit<Model,energy_type>					energy_unit;
	typedef DimensionedUnit<Model,force_type>					force_unit;
	typedef DimensionedUnit<Model,frequency_type>				frequency_unit;
	typedef DimensionedUnit<Model,mass_density_type>			mass_density_unit;
	typedef DimensionedUnit<Model,momentum_type>				momentum_unit;
	typedef DimensionedUnit<Model,power_type>					power_unit;
	typedef DimensionedUnit<Model,pressure_type>				pressure_unit;
	typedef DimensionedUnit<Model,velocity_type>				velocity_unit;
	typedef DimensionedUnit<Model,volume_type>					volume_unit;
	typedef DimensionedUnit<Model,wavenumber_type>				wavenumber_unit;

	typedef DimensionedUnit<Model,capacitance_type>				capacitance_unit;
	typedef DimensionedUnit<Model,conductance_type>				conductance_unit;
	typedef DimensionedUnit<Model,electric_charge_type>			electric_charge_unit;
	typedef DimensionedUnit<Model,electric_potential_type>		electric_potential_unit;
	typedef DimensionedUnit<Model,magnetic_flux_type>			magnetic_flux_unit;
	typedef DimensionedUnit<Model,magnetic_flux_density_type>	magnetic_flux_density_unit;
	typedef DimensionedUnit<Model,inductance_type>				inductance_unit;
	typedef DimensionedUnit<Model,permittivity_type>		    permittivity_unit;
	typedef DimensionedUnit<Model,resistance_type>				resistance_unit;

	typedef DimensionedUnit<Model,catalytic_activity_type>		catalytic_activity_unit;
};

template<class Y,
         class Model>
struct PhysicalQuantityModel
{
    typedef DimensionedQuantity<Y,typename Model::length_unit>                  Length;
    typedef DimensionedQuantity<Y,typename Model::mass_unit>                    Mass;
    typedef DimensionedQuantity<Y,typename Model::time_unit>                    Time;
    typedef DimensionedQuantity<Y,typename Model::current_unit>                 Current;
    typedef DimensionedQuantity<Y,typename Model::temperature_unit>             Temperature;
    typedef DimensionedQuantity<Y,typename Model::quantity_unit>                Quantity;
    typedef DimensionedQuantity<Y,typename Model::intensity_unit>               Intensity;
    typedef DimensionedQuantity<Y,typename Model::angle_unit>             		Angle;
    typedef DimensionedQuantity<Y,typename Model::solid_angle_unit>             SolidAngle;
    
    typedef DimensionedQuantity<Y,typename Model::acceleration_unit>            Acceleration;
    typedef DimensionedQuantity<Y,typename Model::area_unit>                    Area;
    typedef DimensionedQuantity<Y,typename Model::energy_unit>                  Energy;
    typedef DimensionedQuantity<Y,typename Model::force_unit>                   Force;
    typedef DimensionedQuantity<Y,typename Model::frequency_unit>               Frequency;
    typedef DimensionedQuantity<Y,typename Model::mass_density_unit>            MassDensity;
    typedef DimensionedQuantity<Y,typename Model::momentum_unit>                Momentum;
    typedef DimensionedQuantity<Y,typename Model::power_unit>                   Power; 
    typedef DimensionedQuantity<Y,typename Model::pressure_unit>                Pressure; 
    typedef DimensionedQuantity<Y,typename Model::velocity_unit>                Velocity;
    typedef DimensionedQuantity<Y,typename Model::volume_unit>                  Volume;
    typedef DimensionedQuantity<Y,typename Model::wavenumber_unit>              Wavenumber;
    
    typedef DimensionedQuantity<Y,typename Model::capacitance_unit>             Capacitance;
    typedef DimensionedQuantity<Y,typename Model::conductance_unit>             Conductance;
    typedef DimensionedQuantity<Y,typename Model::electric_charge_unit>         ElectricCharge;
    typedef DimensionedQuantity<Y,typename Model::electric_potential_unit>      ElectricPotential;
    typedef DimensionedQuantity<Y,typename Model::inductance_unit>              Inductance;
    typedef DimensionedQuantity<Y,typename Model::magnetic_flux_unit>           MagneticFlux;
    typedef DimensionedQuantity<Y,typename Model::magnetic_flux_density_unit>   MagneticFluxDensity;
    typedef DimensionedQuantity<Y,typename Model::permittivity_unit>            Permittivity;
    typedef DimensionedQuantity<Y,typename Model::resistance_unit>              Resistance;

    typedef DimensionedQuantity<Y,typename Model::catalytic_activity_unit>      CatalyticActivity;
};

} // namespace units

} // namespace mcs

#endif // MCS_PHYSICAL_UNITS_HPP
