// Yet Another Units Library (YANL) - A C++ library for zero-overhead
// dimensional analysis and unit manipulation/conversion.
// 
// Copyright (C) 2003 Matthias Christian Schabel
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#ifndef MCS_SI_UNITS_HPP
#define MCS_SI_UNITS_HPP

#include <string>

#include <mcs/units/dimensioned_quantity.hpp>
#include <mcs/units/physical_units.hpp>

namespace mcs {

namespace units { 

// SI units
struct SIModel :
    public PhysicalUnitModel<SIModel>
{ };

template<> 
struct UnitInfo<SIModel,length_tag> 
{ 
    typedef SIModel  base_type;
    
	static std::string symbol()				{ return "meter"; }
	static std::string abbreviation()		{ return "m"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<SIModel,mass_tag> 
{ 
    typedef SIModel  base_type;
    
	static std::string symbol()				{ return "kilogram"; }
	static std::string abbreviation()		{ return "kg"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<SIModel,time_tag> 
{ 
    typedef SIModel  base_type;
    
	static std::string symbol()				{ return "second"; }
	static std::string abbreviation()		{ return "s"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<SIModel,current_tag> 
{ 
    typedef SIModel  base_type;
    
	static std::string symbol()				{ return "ampere"; }
	static std::string abbreviation()		{ return "A"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<SIModel,temperature_tag> 
{ 
    typedef SIModel  base_type;
    
	static std::string symbol()				{ return "kelvin"; }
	static std::string abbreviation()		{ return "K"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<SIModel,quantity_tag> 
{ 
    typedef SIModel  base_type;
    
	static std::string symbol()				{ return "mole"; }
	static std::string abbreviation()		{ return "mol"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<SIModel,intensity_tag> 
{ 
    typedef SIModel  base_type;

	static std::string symbol()				{ return "candela"; }
	static std::string abbreviation()		{ return "cd"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<SIModel,angle_tag> 
{ 
    typedef SIModel  base_type;
    
	static std::string symbol()				{ return "radian"; }
	static std::string abbreviation()		{ return "rd"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
template<> 
struct UnitInfo<SIModel,solid_angle_tag> 
{ 
    typedef SIModel  base_type;
    
	static std::string symbol()				{ return "steradian"; }
	static std::string abbreviation()		{ return "sr"; }
	
	template<typename Y> static Y toBase()	{ return Y(1); }
};
	
namespace si {

// standard metric prefixes
static const long double	_yocto(1e-24),
							_zepto(1e-21),
							_atto(1e-18),
							_femto(1e-15),
							_pico(1e-12),
							_nano(1e-9),
							_micro(1e-6),
							_milli(1e-3),
							_centi(1e-2),
							_deci(1e-1),
							_deka(1e1),
							_hecto(1e2),
							_kilo(1e3),
							_mega(1e6),
							_giga(1e9),
							_tera(1e12),
							_peta(1e15),
							_exa(1e18),
							_zetta(1e21),
							_yotta(1e24);				
																									 
static const SIModel::length_unit					_meter,
													_meters;
static const SIModel::mass_unit						_kilogram,
													_kilograms;
static const SIModel::time_unit						_second,
													_seconds;
static const SIModel::current_unit					_ampere,
													_amperes;
static const SIModel::temperature_unit				_kelvin,
													_kelvins;
static const SIModel::quantity_unit					_mole,
													_moles;
static const SIModel::intensity_unit				_candela,
													_candelas;
static const SIModel::angle_unit					_radian,
													_radians;
static const SIModel::solid_angle_unit				_steradian,
													_steradians;													
													
static const SIModel::area_unit						_square_meter,
													_square_meters;
static const SIModel::energy_unit					_joule,
													_joules;
static const SIModel::force_unit					_newton,
													_newtons;
static const SIModel::frequency_unit				_hertz;
static const SIModel::power_unit					_watt,
													_watts;
static const SIModel::pressure_unit					_pascal,
													_pascals;
static const SIModel::velocity_unit					_meter_per_second,
													_meters_per_second;
static const SIModel::volume_unit					_cubic_meter,
													_cubic_meters;
														
static const SIModel::capacitance_unit				_farad,
													_farads;
static const SIModel::conductance_unit				_siemen,
													_siemens,
													_mho,
													_mhos;
static const SIModel::electric_charge_unit			_coulomb,
													_coulombs;
static const SIModel::electric_potential_unit		_volt,
													_volts;
static const SIModel::magnetic_flux_unit			_weber,
													_webers;
static const SIModel::magnetic_flux_density_unit	_tesla,
													_teslas;
static const SIModel::inductance_unit				_henry,
													_henrys;
static const SIModel::resistance_unit				_ohm,
													_ohms;

static const SIModel::catalytic_activity_unit		_katal,
													_katals;
													
} // namespace si

template<class Y>
struct SI : 
	public PhysicalQuantityModel<Y,SIModel>
{ };


} // namespace units

} // namespace mcs

#endif // MCS_SI_UNITS_HPP
