// Yet Another Units Library (YANL) - A C++ library for zero-overhead
// dimensional analysis and unit manipulation/conversion.
// 
// Copyright (C) 2003 Matthias Christian Schabel
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#include <iostream>

#include <mcs/units/cgs_units.hpp>
#include <mcs/units/fsf_units.hpp>
#include <mcs/units/si_units.hpp>
#include <mcs/units/dimensioned_io.hpp>

#include <mcs/units/measurement.hpp>

using namespace mcs;
using namespace mcs::units;
using namespace mcs::units::si;

template<class Y,
         class Model>
DimensionedQuantity<Y,DimensionedUnit<Model,energy_type> > 
work(DimensionedQuantity<Y,DimensionedUnit<Model,force_type> > F,
     DimensionedQuantity<Y,DimensionedUnit<Model,length_type> > dx)
{
    return F*dx;
}

template<class Y>
DimensionedQuantity<Y,DimensionedUnit<SIModel,quantity_type> > 
idealGasLaw(DimensionedQuantity<Y,DimensionedUnit<SIModel,pressure_type> > P,
            DimensionedQuantity<Y,DimensionedUnit<SIModel,volume_type> > V,
            DimensionedQuantity<Y,DimensionedUnit<SIModel,temperature_type> > T)
{
    return (P*V/(8.314*(_joules/(_kelvin*_mole))*T));
}

int main()
{
    // dimensioned units
    const double           s1 = 1;
    
    const long             x1 = 2;
    const Rational<4,3>    x2;
    
    //SIModel::length_unit   u1 = _meter;
    //SIModel::time_unit     u2 = _second;
    SIModel::force_unit    u1 = _newton;
    SIModel::energy_unit   u2 = _joule;
    
    //SI<double>::Length     q1(1.0);
    //SI<double>::Time       q2(2.0);
    SI<double>::Force      q1(1.0);
    SI<double>::Energy     q2(2.0);
    
    std::cout << "S1 :    " << s1 << std::endl
              << "X1 :    " << x1 << std::endl
              << "X2 :    " << x2 << std::endl
              << "U1 :    " << u1 << std::endl
              << "U2 :    " << u2 << std::endl
              << "Q1 :    " << q1 << std::endl
              << "Q2 :    " << q2 << std::endl
              << std::endl;
              
    std::cout //<< "U1+U1 : " << u1+u1 << std::endl    // illegal
              //<< "U1-U1 : " << u1-u1 << std::endl    // illegal
              << "U1*U1 : " << u1*u1 << std::endl
              << "U1/U1 : " << u1/u1 << std::endl
              << "U1*U2 : " << u1*u2 << std::endl
              << "U1/U2 : " << u1/u2 << std::endl
              << "U1^X :  " << power<2>(_meter) << std::endl
              << "X1vU1 : " << root<2>(_meter) << std::endl
              << "U1^X2 : " << power<Rational<4,3> >(_meter) << std::endl
              << "X2vU1 : " << root<Rational<4,3> >(_meter) << std::endl
              << std::endl;
    
    std::cout //<< "U1+S1 : " << u1+s1 << std::endl    // illegal
              //<< "S1+U1 : " << s1+u1 << std::endl    // illegal
              //<< "U1-S1 : " << u1-s1 << std::endl    // illegal
              //<< "S1-U1 : " << s1-u1 << std::endl    // illegal
              << "U1*S1 : " << u1*s1 << std::endl
              << "S1*U1 : " << s1*u1 << std::endl
              << "U1/S1 : " << u1/s1 << std::endl
              << "S1/U1 : " << s1/u1 << std::endl
              << std::endl;
              
    std::cout //<< "Q1+S1 : " << q1+s1 << std::endl    // illegal
              //<< "S1+Q1 : " << s1+q1 << std::endl    // illegal
              //<< "Q1-S1 : " << q1-s1 << std::endl    // illegal
              //<< "S1-Q1 : " << s1-q1 << std::endl    // illegal
              << "Q1*S1 : " << q1*s1 << std::endl
              << "S1*Q1 : " << s1*q1 << std::endl
              << "Q1/S1 : " << q1/s1 << std::endl
              << "S1/Q1 : " << s1/q1 << std::endl
              << std::endl;
              
    std::cout //<< "U1+Q1 : " << u1+q1 << std::endl    // illegal
              //<< "Q1+U1 : " << q1+u1 << std::endl    // illegal
              //<< "U1-Q1 : " << u1-q1 << std::endl    // illegal
              //<< "Q1-U1 : " << q1-u1 << std::endl    // illegal
              << "U1*Q1 : " << u1*q1 << std::endl
              << "Q1*U1 : " << q1*u1 << std::endl
              << "U1/Q1 : " << u1/q1 << std::endl
              << "Q1/U1 : " << q1/u1 << std::endl
              << std::endl;
              
    std::cout << "Q1+Q1 : " << q1+q1 << std::endl
              << "Q1-Q1 : " << q1-q1 << std::endl
              << "Q1*Q1 : " << q1*q1 << std::endl
              << "Q1/Q1 : " << q1/q1 << std::endl
              //<< "Q1+Q2 : " << q1+q2 << std::endl    // illegal
              //<< "Q1-Q2 : " << q1-q2 << std::endl    // illegal
              << "Q1*Q2 : " << q1*q2 << std::endl
              << "Q1/Q2 : " << q1/q2 << std::endl
              << "Q1^X1 : " << power<2>(q1) << std::endl
              << "X1vQ1 : " << root<2>(q1) << std::endl
              << "Q1^X2 : " << power<Rational<4,3> >(q1) << std::endl
              << "X2vQ1 : " << root<Rational<4,3> >(q1) << std::endl
              << std::endl;
    
    SI<double>::Length    l1(1.0),
                          l2(1.5);
                        
    std::cout << std::boolalpha
              << (l1 == l2) << std::endl
              << (l1 != l2) << std::endl
              << (l1 <= l2) << std::endl
              << (l1  < l2) << std::endl
              << (l1 >= l2) << std::endl
              << (l1  > l2) << std::endl
              << std::endl;
              
    // unit conversions    
    const double    dimless = (q1/q1);
          
    SI<double>::Volume     vs(1);      
    CGS<double>::Volume    vc(vs);
                        
    std::cout << vs << std::endl
              << vc << std::endl
              << std::endl;
            
    SI<double>::Energy     es(1);      
    CGS<double>::Energy    ec(es);
                        
    std::cout << es << std::endl
              << ec << std::endl
              << std::endl;
            
    SI<double>::Velocity    v1 = 2.0*_meters/_second,
                            v2(CGS<double>::Velocity(7.0));
    
    std::cout << v1 << std::endl
              << v2 << std::endl
              << std::endl;
        
    // combine units library with user-defined type encapsulating error 
    //  propagation
    CGS< Measurement<double> >::Length  u(Measurement<double>(1.0,0.0)),
                                        w(Measurement<double>(4.52,0.02)),
                                        x(Measurement<double>(2.0,0.2)),
                                        y(Measurement<double>(3.0,0.6));
                                        
    std::cout << x+y-w << std::endl;
    std::cout << w*x << std::endl;
    std::cout << x/y << std::endl;
    std::cout << w*y*y/power< Rational<1,2> >(u*x) << std::endl;
    std::cout << std::endl;
    
    SI<double>::Force       F(1.0);
    SI<double>::Length      dx(1.0);
    SI<double>::Energy      E(work(F,dx));
    
    std::cout << F << std::endl
              << dx << std::endl
              << E << std::endl
              << std::endl;
    
    
    // test ideal gas law
    SI<double>::Temperature    T = (273.+37.)*_kelvin;
    SI<double>::Pressure       P = 1.01325e5*_pascals;
    SI<double>::Length         r = 0.5e-6*_meters;
    SI<double>::Volume         V = (4.0/3.0)*3.141592*power<3>(r);
    SI<double>::Quantity       n(idealGasLaw(P,V,T));
    
    std::cout << r << std::endl
              << P << std::endl
              << V << std::endl
              << T << std::endl
              << n << std::endl
              << std::endl;
              
    // test furlong/stone/fortnight units
    FSF<double>::Length fl(1);  // 1 furlong
    FSF<double>::Time   ft(1);  // 1 fortnight
    
    CGS<double>::Velocity   flpft(fl/ft);
    
    std::cout << fl << std::endl
              << ft << std::endl
              << fl/ft << std::endl
              << flpft << std::endl;
    
    return 0;
}
